package jclusim;

import java.io.File;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.Iterator;

import jclusim.base.Experiment;
import jclusim.base.IExperimentSettings;
import jclusim.observer.EnvironmentObserver;
import jclusim.observer.ExperimentObserver;
import jclusim.observer.StatisticsObserver;
import jclusim.ui.JClusimFrame;

public class JClusim implements Runnable {

    public static final String INTERACTIVE_PROPERTY_NAME = "jclusim.interactive";
    public static final String EXPERIMENT_PROPERTY_NAME = "jclusim.experiment";

    private static final File RESULTS_DIR = new File("Results");

    private static final String DEFAULT_EXPERIMENT = "jclusim.experiments.simple.SimpleSettings";
    private static final String DEFAULT_INTERACTIVE = Boolean.TRUE.toString();

    private Experiment experiment;
    private boolean interactive;
    private ArrayList<ExperimentObserver> expObservers;
    private File experimentDir;
    private JClusimFrame jcf;

    @SuppressWarnings("unchecked")
    public static void main(String[] args) {

	String experimentClassName = System.getProperty(
		EXPERIMENT_PROPERTY_NAME, DEFAULT_EXPERIMENT);
	Boolean interactive = Boolean.valueOf(System.getProperty(
		INTERACTIVE_PROPERTY_NAME, DEFAULT_INTERACTIVE));

	try {
	    Class<? extends IExperimentSettings> settingsClass = (Class<? extends IExperimentSettings>) Class
		    .forName(experimentClassName);
	    IExperimentSettings settings = settingsClass.newInstance();
	    JClusim sim = new JClusim(settings, interactive);
	    sim.run();
	} catch (ClassNotFoundException e) {
	    System.err.println("Settings class not found: " + e.getMessage());
	} catch (IllegalAccessException e) {
	    System.err.println("Cannot access settings class: "
		    + e.getMessage());
	} catch (InstantiationException e) {
	    System.err.println("Cannot instantiate settings class: "
		    + e.getMessage());
	}
    }

    public JClusim(IExperimentSettings settings, boolean interactive) {

	this.interactive = interactive;

	experiment = new Experiment(settings);

	expObservers = new ArrayList<ExperimentObserver>();
	expObservers.add(new EnvironmentObserver());
	expObservers.add(new StatisticsObserver());

	if (interactive) {
	    jcf = new JClusimFrame(expObservers);
	    jcf.setVisible(true);
	    jcf.pack();
	}

    }

    public synchronized void run() {
	reset();
	if (!interactive) {
	    createExperimentResultDirectory();
	    takeObserversSnapshot();
	}

	experiment.startExperiment();

	if (!interactive) {
	    takeObserversSnapshot();
	    dumpObserversState();
	    summarizeObserversState();
	}
    }

    private void createExperimentResultDirectory() {
	String date = new SimpleDateFormat("yyyyMMdd_HHmmss")
		.format(new Date());
	String experimentName = experiment.getSettings().getClass()
		.getSimpleName();

	experimentDir = new File(RESULTS_DIR, date + "_" + experimentName);
	experimentDir.mkdirs();
    }

    private void dumpObserversState() {
	for (Iterator<ExperimentObserver> it = expObservers.iterator(); it
		.hasNext();) {
	    it.next().dumpState(experimentDir);
	}
    }

    private void summarizeObserversState() {
	for (Iterator<ExperimentObserver> it = expObservers.iterator(); it
		.hasNext();) {
	    System.out.println(it.next().summarizeState());
	}
    }

    private void takeObserversSnapshot() {
	for (Iterator<ExperimentObserver> it = expObservers.iterator(); it
		.hasNext();) {
	    it.next().takeSnaphot(experimentDir);
	}
    }

    private void reset() {
	experiment.deleteObservers();
	experiment.reset();

	for (Iterator<ExperimentObserver> it = expObservers.iterator(); it
		.hasNext();) {
	    ExperimentObserver obs = it.next();
	    experiment.addObserver(obs);
	    obs.resetState(experiment);
	}
	experiment.notifyObservers();
    }

}