package jclusim.observer;

import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.geom.Rectangle2D;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;

import jclusim.base.Experiment;
import jclusim.observer.metric.Activity;
import jclusim.observer.metric.Homogeneity;
import jclusim.observer.metric.IMetric;
import jclusim.util.Similarity;

import org.jfree.chart.ChartFactory;
import org.jfree.chart.JFreeChart;
import org.jfree.chart.axis.NumberAxis;
import org.jfree.chart.plot.PlotOrientation;
import org.jfree.chart.plot.XYPlot;
import org.jfree.chart.renderer.xy.XYLineAndShapeRenderer;
import org.jfree.data.statistics.Statistics;
import org.jfree.data.xy.XYDataItem;
import org.jfree.data.xy.XYSeries;
import org.jfree.data.xy.XYSeriesCollection;

public class StatisticsObserver extends ExperimentObserver {

    private int METRIC_RESOLUTION = 2000;

    private static final int IMAGE_HEIGHT = 300;
    private static final int IMAGE_WIDTH = 640;

    private static final IMetric[] metrics = { new Activity(),
	    new Homogeneity(Similarity.COSINUS),
	    new Homogeneity(Similarity.EUCLIDEAN) };

    private XYSeriesCollection dataset;
    private JFreeChart chart;

    public StatisticsObserver() {

	this.dataset = new XYSeriesCollection();

	for (int i = 0; i < metrics.length; i++) {
	    XYSeries series = new XYSeries(metrics[i].toString(), false, false);
	    this.dataset.addSeries(series);
	}

	chart = ChartFactory.createXYLineChart("Plot of Metrics", // chart
		// title
		"Iteration", // x axis label
		"Value of metric", // y axis label
		dataset, // data
		PlotOrientation.VERTICAL, true, // include legend
		true, // tooltips
		false // urls
		);

	chart.setBackgroundPaint(Color.white);
	chart.setTitle("Metrics Values");

	XYPlot plot = chart.getXYPlot();
	plot.setBackgroundPaint(Color.lightGray);
	plot.setDomainGridlinePaint(Color.white);
	plot.setRangeGridlinePaint(Color.white);

	XYLineAndShapeRenderer renderer = new XYLineAndShapeRenderer(true,
		false);
	plot.setRenderer(renderer);

	NumberAxis rangeAxis = (NumberAxis) plot.getRangeAxis();
	rangeAxis.setAutoRange(true);

    }

    @SuppressWarnings("unchecked")
    @Override
    public synchronized void resetState(Experiment exp) {
	for (Iterator<XYSeries> it = dataset.getSeries().iterator(); it
		.hasNext();) {
	    it.next().clear();
	}
    }

    @Override
    public synchronized void updateState(Experiment exp) {
	int n = exp.getIterationNumber();

	for (int i = 0; i < metrics.length; i++) {
	    IMetric metric = metrics[i];
	    dataset.getSeries(metric.toString()).add(n,
		    metric.computeValue(exp));
	}

    }

    @Override
    public String toString() {
	return "Statistics";
    }

    @Override
    public synchronized void renderState(Graphics2D g2d) {
	chart
		.draw(g2d, new Rectangle2D.Double(0, 0, IMAGE_WIDTH,
			IMAGE_HEIGHT));
    }

    @Override
    public int getCanevasHeight() {
	return IMAGE_HEIGHT;
    }

    @Override
    public int getCanevasWidth() {
	return IMAGE_WIDTH;
    }

    @Override
    public int getRefreshRate() {
	return METRIC_RESOLUTION;
    }

    @SuppressWarnings("unchecked")
    @Override
    public String summarizeState() {
	StringBuffer sb = new StringBuffer();
	for (Iterator<XYSeries> it = dataset.getSeries().iterator(); it
		.hasNext();) {
	    XYSeries serie = it.next();
	    List<Double> yValues = gatherYValuesFromXYSerie(serie);
	    sb.append('\n');
	    sb.append(serie.getKey() + ":");
	    sb.append('\n');
	    sb.append("\tBegin: " + yValues.get(0));
	    sb.append('\n');
	    sb.append("\tEnd: " + yValues.get(yValues.size() - 1));
	    sb.append('\n');
	    sb.append("\tMean: " + Statistics.calculateMean(yValues));
	    sb.append('\n');
	    sb.append("\tMedian: " + Statistics.calculateMedian(yValues));
	    sb.append('\n');
	    Collections.sort(yValues);
	    sb.append("\tMin: " + yValues.get(0));
	    sb.append('\n');
	    sb.append("\tMax: " + yValues.get(yValues.size() - 1));
	    sb.append('\n');
	}
	return sb.toString();
    }

    @SuppressWarnings("unchecked")
    private static List<Double> gatherYValuesFromXYSerie(XYSeries serie) {
	List<Double> yValues = new ArrayList<Double>(serie.getItemCount());

	for (Iterator<XYDataItem> itSerie = serie.getItems().iterator(); itSerie
		.hasNext();) {
	    XYDataItem dItem = itSerie.next();
	    yValues.add(dItem.getYValue());
	}

	return yValues;
    }

    @SuppressWarnings("unchecked")
    @Override
    public void dumpState(File folder) {
	for (Iterator<XYSeries> it = dataset.getSeries().iterator(); it
		.hasNext();) {
	    try {
		dumpSerie(folder, it.next());
	    } catch (IOException ex) {
		ex.printStackTrace();
	    }
	}
    }

    @SuppressWarnings("unchecked")
    private void dumpSerie(File folder, XYSeries serie) throws IOException {
	String fileName = serie.getKey() + ".dat";
	File dumpFile = new File(folder, fileName);

	FileWriter fw = new FileWriter(dumpFile);
	BufferedWriter bw = new BufferedWriter(fw);
	for (Iterator<XYDataItem> it = serie.getItems().iterator(); it
		.hasNext();) {
	    XYDataItem item = it.next();
	    bw.write(item.getXValue() + "\t" + item.getYValue());
	    if (it.hasNext()) {
		bw.newLine();
	    }
	}
	bw.flush();
	bw.close();

    }

}
